//
//  BITSynth.cpp
//  Bitalino
//
//  Created by James on 23/11/2017.
//
//

#include "BITSynth.hpp"

BITSynth::BITSynth ( void ) : ThreadedProcess ()
{
    inputData(0.0);
    currentState.store(IDLE);
    recording.store(false);
    selectingData.store(false);
    redraw.store(false);
    transport.store(0.0);
    drawState.store(DRAW_IDLE);
}

BITSynth::~BITSynth ( void )
{
    
}

void BITSynth::setup( ofRectangle posAndSize )
{
    this->posAndSize = posAndSize;
    // Set up visualization
    //rv.setup(posAndSize, 1024);
    resize(posAndSize);
    rv.setGuiHidden(true);
    
    // Set model creation callback: (not sync)
    rapidMixThread.modelCreationCallback = [&] ( void )
    { // This callback is just in case xmm is used (to change init stuff, probably change to diff setup method?)
        rapidmix::xmmConfig xcfg;
        xcfg.relativeRegularization = 0.005;
        xcfg.absoluteRegularization = 0.005;
        xcfg.states = 12;
        xcfg.likelihoodWindow = 10;
        selectedTraining* t = new selectedTraining();//xcfg);
        t->setNumHiddenNodes(18);
        //t->setNumHiddenLayers(2);
        //t->setNumEpochs(50000);
        return t;
    };
    // Set threaded train callback: (not sync)
    rapidMixThread.threadedTrainCallback = [&] (selectedTraining* t )
    {
        printf("Started training in thread... \n");
        if (!trainingData.trainingSet.empty() && trainingData.trainingSet[0].elements.size() > 1)
            if (!t->train(trainingData))
                printf("Training Failed\n");
    };
    // Set callback for when model is done training (called in rapidMixThread.update method)
    rapidMixThread.doneTrainingCallback = [&] ( void )
    {
        sendRunDataToOFX();
        transport.store(0.0);
        initialTrainingDone = true;
        //trainingText = "New Model Trained";
        //printf("New Model Trained\n");
    };
    // Set buttonPressedCallback for rapidBitalino
    rapidBitalino.buttonPressedCallback = [&] ( void )
    { // Is called in sync with Data thread
        longSelectionTimer.reset();
        switch (currentState.load())
        { // Run update for state
            case IDLE:
                printf("Started \n");
                generateFirstData();
                currentState.store(PLAYING);
                break;
            case PERFORMING:
            {
                if (recording.load())
                {
                    printf("Playing recorded gestures \n");
                    recording.store(false);
                    currentState.store(PLAYING);
                } else {
                    printf("Started recording \n");
                    recording.store(true);
                }
                break;
            }
            case PLAYING:
            {
                if (!selectingData.load())
                {
                    printf("Selecting data for example \n");
                    trainingData.startRecording("NewPhrase");
                    selectingData.store(true);
                }
                break;
            }
            default:
                break;
        }
    };

    // Set buttonReasedCallback for rapidBitalino
    rapidBitalino.buttonReleasedCallback = [&] ( void )
    { // Is called in sync with Data thread
        switch (currentState.load())
        { // Run update for state
            case PLAYING:
            {
                if (selectingData.load())
                {
                    if (longSelectionTimer.getTimeElapsed() > longPress)
                    {
                        printf("Recorded example \n");
                        selectingData.store(false);
                    } else {
                        printf("Throwing away last training sample, press was for stopping \n");
                        trainingData.trainingSet.pop_back();
                        
                        if (trainingData.trainingSet.size() >= 1)
                        {
                            //if (dataIsAbove) TODO test if data is not to flat
                            printf("Training on selected data \n");
                            rapidMixThread.train();
                            currentState.store(PERFORMING);
                        } else {
                            printf("Reinitializing");
                            loop.clear();
                            generateFirstData();
                            sendLoopDataToOFX();
                        }
                        selectingData.store(false);
                    }
                }
                break;
            }
            default:
                break;
        }
    };

    // Set up training thread
    rapidMixThread.setup();
    
    // Set up BITalino accessor + thread
    rapidBitalino.setup( 10 ); // downsampling 10 of the data (TODO experiment with different settings with bayes)
    
    // Set up the ringbuffer for controldata which is pushed in to the audio thread
    controlDataBuffer.setup(100); // Leave room for 10 control data to sync over to audio thread
    // ^ TODO try highter control data storage room with usleep implemented in data thread
    
    toOFXVis.setup(1000); // Leave room for 1000 data to sync over to ofx side
    
    toOFXRunVis.setup(1000); // Leave room for 1000 data to sync over to ofx side
    
    // Start the thread
    startThread();
}

void BITSynth::setState ( State state )
{
    this->currentState.store(state);
}

void BITSynth::inputData( double input )
{ // This being atomic can negate the need for a ringbuffer...
    // TODO think of what will be the solution of input here
    //currentInput.store(0.0);
}

void BITSynth::update ( void )
{
    if (redraw.load()) // replace with safe condition / lock?
    {
        rv.reset();
        uint32_t itemsToRead = toOFXRunVis.items_available_for_read();
        drawType t = drawState.load();
        if (t == DRAW_RUN)
        {
            itemsToRead = toOFXRunVis.items_available_for_read();
        } else if (t == DRAW_PLAYBACK) {
            itemsToRead = toOFXVis.items_available_for_read();
        }
        
        if (itemsToRead)
        {
            rv.setHistorySize(itemsToRead);
            SynthControlData data[itemsToRead];
            if (t == DRAW_RUN)
            {
                toOFXRunVis.pop(data, itemsToRead);
            } else if (t == DRAW_PLAYBACK) {
                toOFXVis.pop(data, itemsToRead);
            }
            
            for (uint32_t i=0; i < itemsToRead; ++i)
            {
                std::vector<double> dVect = data[i].getValuesAsVector();
                for (uint32_t param=0; param < BIT::NUM_SYNTH_PARAMS; ++param)
                {
                    std::vector<double> sVect = {dVect[param]};
                    rv.addData("/"+std::string(BIT::ParamNames[param]), sVect);
                }
            }
        }
        redraw.store(false);
    }

    rv.update();
}

void BITSynth::draw ( void )
{
    rv.draw();
    
    double xPosTransport = posAndSize.x + posAndSize.width * transport.load();
    ofDrawLine(xPosTransport, posAndSize.y, xPosTransport, posAndSize.y + posAndSize.height);
    
    rapidBitalino.draw();
    drawTextLabel(stateStrings[currentState.load()], ofVec2f(posAndSize.x, posAndSize.y+21), ofColor(24, 219, 92), ofColor(10, 10, 10), TextAlignment::LEFT, false );
}

void BITSynth::resize ( ofRectangle posAndSize )
{
    this->posAndSize = posAndSize;
    float heightOffset = 48.0;
    
    
    rv.setPosAndSize(ofRectangle(posAndSize.x, posAndSize.y + heightOffset, posAndSize.width, posAndSize.height-heightOffset));
}

void BITSynth::stop( void )
{
    // Stop all asynchronous threads
    stopThread();
    rapidMixThread.stopThread();
    rapidBitalino.stop();
}

void BITSynth::audioOut( float *output, int bufferSize, int nChannels )
{ // Audio thread sync callback
    for (uint32_t i = 0; i < bufferSize; ++i)
    {
        if (controlDataBuffer.items_available_for_read())
            controlDataBuffer.pop(&audioControlData, 1);
        
        SVF.setCutoff(dFilt1.lopass(110 + fmin(fabs(audioControlData[BIT::CUTOFF]), 0.95) * 5000, 0.001));
        SVF.setResonance(0.1 + fmin(fabs(audioControlData[BIT::RESONANCE]), 0.9));
        
        double modulator = VCO2.sinewave(fabs(20 + audioControlData[BIT::MODULATOR]  * 1000.0)) * 50;
        double carrier = VCO1.sinewave(fabs(audioControlData[BIT::CARRIER] + modulator));
        double filtered = SVF.play(carrier, 1.0, 0, 0, 0) * 0.25;
        double delay1 = DL1.dl(filtered, dFilt2.lopass(fmin(fabs(audioControlData[BIT::SIZE_A]) * 88200, 88199), 0.01),
                                              fabs(audioControlData[BIT::FB_A]));
        double delay2 = DL2.dl(filtered, dFilt3.lopass(fmin(fabs(audioControlData[BIT::SIZE_B]) * 88200, 88199), 0.01),
                                              fabs(audioControlData[BIT::FB_B]));
        
        output[i * nChannels] = filtered * 0.75 + delay1 * 0.125 + delay2 * 0.0125;
        output[i * nChannels + 1] = filtered * 0.75 + delay1 * 0.0125 + delay2 * 0.125;
    }
}

void BITSynth::mainThreadCallback( void )
{ // Process thread sync callback
    // Get input data (BITalino):
    std::vector<double> inputData = rapidBitalino.update(); // Update method uses ringbuffer and is threadsafe
    //maxValueEMG = max(inputData[0], maxValueEMG);
    printf("input = %f\n", inputData[0]);
    State curState = currentState.load();
    bool newState = (curState != previousState);
    
    // TODO this can be cleaned up.. could put the states in classes/structs
    if (newState)
    {
        dataPlayTimer.reset();
        longSelectionTimer.reset();
        switch (curState)
        { // Run inits for states
            case PERFORMING:
                loop.clear();
                break;
            case PLAYING:
                sendLoopDataToOFX();
                loopPosition = loop.begin();
                trainingData.trainingSet.clear();
                break;
            default:
                break;
        }
    } else
    switch (curState)
    { // Run update for state
        case PERFORMING:
        {
            if (dataPlayTimer.getTimeElapsed() > controlRate && initialTrainingDone)
            {
                transport.store(inputData[0]/maxValueEMG);
                SynthControlData ctrlData(rapidMixThread.model->run(inputData));
                controlDataBuffer.push(&ctrlData, 1);
                currentControlData = ctrlData;
                if (recording.load())
                {
                    loop.push_back(ctrlData);
                }
                dataPlayTimer.reset();
            }
            break;
        }
        case PLAYING:
        {   // Set loopPosition to begin in state change?
            
            if (!loop.empty() && dataPlayTimer.getTimeElapsed() > controlRate)
            {
                SynthControlData ctrlData = *loopPosition;
                controlDataBuffer.push(&ctrlData, 1);
                currentControlData = ctrlData;
                if (++loopPosition == loop.end())
                {
                    loopPosition = loop.begin();
                }
                
                transport.store(static_cast<double>(loopPosition - loop.begin()) / loop.size());
                
                if (selectingData.load() && oldInputData != inputData[0])
                {
                    std::vector<double> outputTraining = ctrlData.getValuesAsVector();
                    //printf("%f, %f\n", inputData[0], outputTraining[0], outputTraining[1], outputTraining[2], outputTraining[3]);
                    trainingData.addElement(inputData, outputTraining);
                    oldInputData = inputData[0];
                }
                dataPlayTimer.reset();
            }
            break;
        }
        default:
            break;
    }
    
    // TODO add usleep (fine tune for data synchronization)
    rapidMixThread.update();
    previousState = curState;
}

void BITSynth::generateFirstData ( void )
{
    uint8_t midiStart = 55;
    uint32_t doremiSize = 2;
    //SynthControlData data[doremiSize * BIT::majorScaleSize]; // Doremi
    // First generate doremi ladder
    for (uint32_t i = 0; i < doremiSize; ++i)
    {
        for (uint32_t step = 0; step < BIT::majorScaleSize; ++step)
        {
            double frequency = MTOF(midiStart + ( i * 12 ) + BIT::majorScale[step]);
            SynthControlData ctrlData;
            ctrlData.values[BIT::CARRIER] = frequency;
            for (uint16_t x = 0; x < BIT::NUM_SYNTH_PARAMS; ++x)
            {
                if (x != static_cast<uint8_t>(BIT::CARRIER))
                    ctrlData.values[x] = ((double)rand()/RAND_MAX);
            }

            for (uint32_t y = 0; y < samplesPerExample; ++y)
                loop.push_back(ctrlData);
            
            //data[i * step] = ctrlData;
        }
    }
}

void BITSynth::generateFirstRandomData ( void )
{ // TODO TEMPORARY
    for (uint32_t i = 0; i < 10; ++i)
    {
        SynthControlData ctrlData;
        for (uint16_t x = 0; x < BIT::NUM_SYNTH_PARAMS; ++x)
        {
            ctrlData.values[x] = ((double)rand()/RAND_MAX);
        }
        for (uint32_t y = 0; y < samplesPerExample; ++y)
            loop.push_back(ctrlData);
    }
}

void BITSynth::sendRunDataToOFX ( void )
{
    toOFXRunVis.reset();
    for (double i=0.0; i<maxValueEMG; i+= 1)
    {
        std::vector<double> t(rapidBitalino.numOutputs, i);
        SynthControlData data(rapidMixThread.model->run(t));
        std::vector<double> d = data.getValuesAsVector();
        for (int p = 0; p < d.size(); ++p)
        {
            printf("%f: %s, %f \n", i, BIT::ParamNames[p], d[p]);
        }
        toOFXRunVis.push(&data, 1);
    }
    
    drawState.store(DRAW_RUN);
    redraw.store(true);
}

void BITSynth::sendLoopDataToOFX ( void )
{
    toOFXVis.reset(); // Reset read and write pointers
    if (toOFXVis.items_available_for_write() < loop.size())
    {
        // TODO: test if resizing is threadsafe in this context... or implement atomics for it
        toOFXVis.setSize(loop.size() + 1); // This can be called here because not audio (but will block for a bit)
    }
    
    toOFXVis.push(&loop[0],  loop.size());
    drawState.store(DRAW_PLAYBACK);
    redraw.store(true);
}
