//
//  BITSynth.hpp
//  Bitalino
//
//  Created by James on 23/11/2017.
//
//

#ifndef BITSynth_hpp
#define BITSynth_hpp

#include <stdio.h>
#include <vector>
#include <atomic>
#include <algorithm>
#include <iterator>
#include <mutex>

// Rapid Mix and other stuff
#include "ThreadedProcess.h"
#include "rapidmix.h"
#include "RapidMixThread.h"
#include "RapidBitalino.h"
#include "RingBuffer.hpp"
#include "ofxMaxim.h"
#include "ofMain.h"
#include "Timer.h"
#include "RapidVisualization.hpp"
#include "SigTools.hpp"

typedef rapidmix::staticRegression selectedTraining;

namespace BIT
{
    enum SynthControl
    {
        CARRIER,
        MODULATOR,
        AMP,
        CUTOFF,
        RESONANCE,
        SIZE_A,
        SIZE_B,
        FB_A,
        FB_B,
        NUM_SYNTH_PARAMS // ALWAYS AT END, REMEMBER TO ADD YOUR PARAMETER AS A STRING BELOW ASWELL
    };
    
    inline const uint8_t majorScaleSize = 7;
    
    inline const uint8_t majorScale[majorScaleSize]
    {
        0, 2, 4, 5, 7, 9, 11
    };

    // C++17 only: (to use below c++17 put it in an instance of some <datatype>)
    inline const char* ParamNames[NUM_SYNTH_PARAMS] = {
        "1 Carrier",
        "2 Modulator",
        "3 Amp",
        "4 Cutoff",
        "5 Resonance",
        "6 SizeA",
        "7 SizeB",
        "8 FeedbackA",
        "9 FeedbackB"
    };

}

struct SynthControlData
{
    double values[BIT::NUM_SYNTH_PARAMS];
    SynthControlData ( void )
    {
    }
    
    SynthControlData ( std::vector<double> data )
    {
        setValues (data);
    }
    
    double& operator[] (const BIT::SynthControl index)
    {
        return values[static_cast<uint32_t>(index)];
    }
    
    int32_t setValues ( std::vector<double>& data )
    { // Returns number of elements which exceeds limit of synth param size
        std::copy(data.begin(), data.begin() + BIT::NUM_SYNTH_PARAMS, values);
        return data.size() - BIT::NUM_SYNTH_PARAMS;
    }
    
    std::vector<double> getValuesAsVector ( void )
    {
        return std::vector<double>(std::begin(values), std::end(values));
    }
};

class BITSynth : public ThreadedProcess
{
public:
    enum State {
        TRAINING,
        PERFORMING,
        PLAYING,
        IDLE,
        NUMBER_OF_STATES
    };
    const char* stateStrings[NUMBER_OF_STATES]
    {
        "Training",
        "Performing",
        "Playing",
        "Idle"
    };

    BITSynth ( void );
    ~BITSynth ( void );
    
    void setup ( ofRectangle posAndSize );
    
    void setState ( State state );
    void inputData ( double input );

    void update ( void );
    void draw ( void );
    void resize ( ofRectangle posAndSize );
    void stop ( void );
    
    void audioOut(float * output, int bufferSize, int nChannels);
    
protected:
    void mainThreadCallback ( void );
    void generateFirstData ( void );
    void generateFirstRandomData ( void );
    void sendLoopDataToOFX ( void );
    void sendRunDataToOFX ( void );
    
private:
    // ofMaxim audio stuff
    maxiOsc VCO1, VCO2;
    maxiSVF SVF;
    
    maxiFractionalDelay DL1, DL2;
    // For smoothing parameter changes:
    maxiFilter dFilt1, dFilt2, dFilt3;
    
    // Rapidmix stuff
    // TODO: turn the threaded base class that is inherited from in to a
    // class which you can add callbacks to so these arent weirdly balanced
    // between the two.
    RapidMixThread<selectedTraining> rapidMixThread;
    rapidmix::trainingData trainingData;
    
    //std::string trainingText = "idle";
    bool gatheringTrainingData = false;
    
    // BITalino thread, TODO: Read above
    RapidBitalino rapidBitalino;
    
    // Other stuff
    uint32_t sampleRate = 44100;
    uint32_t controlRate = 50; // 50ms
    uint32_t samplesPerExample = 20;
    
    // DEBUGDEBUG
    double oldInputData = 0.0;
    //--
    
    
    std::atomic<State> currentState;
    State previousState; // (synced with Data thread)
    
    std::atomic<bool> recording;
    std::atomic<bool> selectingData;
    
    // Data thread synced
    SynthControlData currentControlData;
    std::vector<SynthControlData> loop;
    std::vector<SynthControlData>::iterator loopPosition;
    bool initialTrainingDone = false;
    
    // Timers
    Timer dataPlayTimer;
    Timer longSelectionTimer;
    
    uint32_t longPress = 400; // 200ms
    
    // Audio thread synced
    SynthControlData audioControlData;
    
    // From Data thread to Audio thread
    RingBuffer<SynthControlData> controlDataBuffer;
    RingBuffer<SynthControlData> toOFXVis;
    RingBuffer<SynthControlData> toOFXRunVis;
    
    // OFX stuff
    RapidVisualization rv;
    ofRectangle posAndSize;
    enum drawType {
        DRAW_IDLE,
        DRAW_RUN,
        DRAW_PLAYBACK
    };
    
    double maxValueEMG = 600.0;
    
    // Data to OFX threadsafe
    std::atomic<bool> redraw;
    std::atomic<drawType> drawState;
    std::atomic<double> transport;
    
};

#endif /* BITSynth_hpp */
